#include <libxml/xmlwriter.h>
#include "cloog_util.h"
#include "xml_AST.h"

#ifndef POL_INTEGER
#define POL_INTEGER	0
#endif

#define MAXRAYS    (POL_NO_DUAL | POL_INTEGER)

static int read_int(FILE *in)
{
    char s[1024];
    int i = -1;

    while (fgets(s, sizeof(s), stdin)) {
	if (*s == '#')
	    continue;
	if (sscanf(s, "%d", &i) == 1)
	    break;
    }
    return i;
}

static char *read_str(FILE *in)
{
    char s[1024];
    char str[1024];
    int i = -1;

    while (fgets(s, sizeof(s), stdin)) {
	if (*s == '#')
	    continue;
	if (sscanf(s, "%s", &str) == 1)
	    break;
    }
    return strdup(str);
}

static void add_scat(cloog_program_builder& pb, int nparam, int dim, 
		     Polyhedron *domain, int offset)
{
    int ddim = domain->Dimension - nparam;
    Matrix *M = Matrix_Alloc(dim+1, 1+(dim+1)+ddim+nparam+1);
    for (int i = 0; i < dim; ++i) {
	value_set_si(M->p[i][1+i], 1);
	value_set_si(M->p[i][1+(dim+1)+i], -1);
    }
    value_set_si(M->p[dim][1+dim], 1);
    value_set_si(M->p[dim][1+(dim+1)+ddim+nparam], -offset);
    Polyhedron *scat = Constraints2Polyhedron(M, POL_NO_DUAL);
    
    pb.add_scattering(scat);
}

int main(int argc, char * argv[])
{
    Matrix *M;
    Polyhedron *C;
    xmlTextWriterPtr writer;
    xmlDocPtr doc;
    int rc;
    seq<pdg::parameter>  params;

    LIBXML_TEST_VERSION

    int dim = read_int(stdin);
    const char **iter_name = Read_ParamNames(stdin, dim);
    int nparam = read_int(stdin);
    const char **param_name = Read_ParamNames(stdin, nparam);

    M = Matrix_Read();
    C = Constraints2Polyhedron(M, MAXRAYS);
    Matrix_Free(M);

    cloog_program_builder pb(C);

    int n = read_int(stdin);
    for (int i = 0; i < n; ++i) {
	char *node = read_str(stdin);
	Polyhedron *D;
	fprintf(stderr, "%s\n", node);

	M = Matrix_Read();
	D = Constraints2Polyhedron(M, MAXRAYS);
	Matrix_Free(M);

	pb.add_domain(D, i, node);
	add_scat(pb, nparam, dim, D, 3*i+1);

	int n_in = read_int(stdin);
	for (int j = 0; j < n_in; ++j) {
	    char *port = read_str(stdin);

	    M = Matrix_Read();
	    D = Constraints2Polyhedron(M, MAXRAYS);
	    Matrix_Free(M);

	    pb.add_domain(D, i, port);
	    add_scat(pb, nparam, dim, D, 3*i);
	}

	int n_out = read_int(stdin);
	for (int j = 0; j < n_out; ++j) {
	    char *port = read_str(stdin);

	    M = Matrix_Read();
	    D = Constraints2Polyhedron(M, MAXRAYS);
	    Matrix_Free(M);

	    pb.add_domain(D, i, port);
	    add_scat(pb, nparam, dim, D, 3*i+2);
	}
    }

    for (int i = 0; i < nparam; ++i) {
	pdg::parameter *p = new pdg::parameter;
	p->name = new str(param_name[i]);
	params.push_back(p);
    }
    ++dim;
    pb.set_names(params, dim);
    cloog_program_scatter(pb.p, pb.scatterings);

    writer = xmlNewTextWriterDoc(&doc, 0);
    assert(writer);

    rc = xmlTextWriterStartDocument(writer, NULL, NULL, NULL);
    assert(rc >= 0);

    rc = xmlTextWriterStartDTD(writer, BAD_CAST "ast",
			    BAD_CAST "-//LIACS//DTD ESPAM 1//EN",
			    BAD_CAST "http://www.liacs.nl/~cserc/dtd/espam_1.dtd");
    assert(rc >= 0);
    rc = xmlTextWriterEndDTD(writer);
    assert(rc >= 0);

    writeAST(writer, pb.p, dim);

    rc = xmlTextWriterEndDocument(writer);
    assert(rc >= 0);

    xmlFreeTextWriter(writer);

    rc = xmlDocFormatDump(stdout, doc, 1);
    assert(rc >= 0);
}
