#ifndef SIZE_H
#define SIZE_H

#include "config.h"
#ifdef HAVE_BERNSTEIN
#include <bernstein/piecewise_lst.h>
#endif
#include <barvinok/barvinok.h>
#include "pdg.h"

namespace size {

struct enumerator {
    enum type {
	fixed,
	polylib,
	ginac,
    } type;
    int v;
    evalue *EP;
    pdg::PDG *pdg;
#ifdef HAVE_BERNSTEIN
    bernstein::piecewise_lst *pl;
    enumerator(bernstein::piecewise_lst *pl, pdg::PDG *pdg) : 
		    type(ginac), pl(pl), pdg(pdg) {}
#endif
    enumerator(int v) : type(fixed), v(v) {}
    enumerator(evalue *e, pdg::PDG *pdg) : type(polylib), EP(e), pdg(pdg) {}
    operator pdg::enumerator *() const;
    /* evaluate the enumerator at the values stored in the parameters */
    integer *evaluate() const;
    ~enumerator() {
	if (type == polylib) {
	    free_evalue_refs(EP);
	    free(EP);
	}
#ifdef HAVE_BERNSTEIN
	if (type == ginac)
	    delete pl;
#endif
    }
};

const int size_simulate = 1;
const int size_bernstein = 2;

struct size_options {
    barvinok_options	*barvinok;
    int			 flags;
};

size::enumerator *selfloop_size(pdg::PDG *pdg, Polyhedron *dep, 
			       int n_write, int n_read, Polyhedron *ctx,
			       size_options *options);

}

#endif
