#include "suif_inc.h"
#include <stdlib.h>
#include <assert.h>
#include "domain.h"
#include "build.h"

#include "yaml.h"
#include "pdg.h"
#include <cloog/cloog.h>
#include "cloog_util.h"

using pdg::PDG;

static CloogProgram * program_read(char* in, int dump);
static void program_dump(PDG* pdg);

#define ARRAY_SIZE(a) (sizeof(a)/sizeof(*a))

static CloogProgram *program;
static CloogOptions *options;

static int oldstyle;

static void do_proc(tree_proc * tp)
{
    proc_sym * psym = tp->proc();
    fill_in_access(tp);
    construct_domains(tp, 0, oldstyle, NULL);
    build(tp, program, options);
}

int
main(int argc, char * argv[])
{
    int dump = 0;
    char *domains = NULL;

    options = cloog_options_malloc() ;
    options->sh = 1;
    cmd_line_option my_options[] = {
	{ CLO_NOARG,	"-dump",	"",	&dump },
	{ CLO_STRING,   "-domains",	"", 	&domains },
	{ CLO_INT,	"-f",		"1", 	&options->f },
	{ CLO_INT,	"-strides",	"0", 	&options->strides },
	{ CLO_NOARG,  	"-oldstyle",	"", 	&oldstyle },
    };
    
    start_suif(argc, argv);
    init_dead_code();
    init_unused();
    parse_cmd_line(argc, argv, my_options, ARRAY_SIZE(my_options));
    init(argc, argv);

    program = program_read(domains, dump);
    if (program == NULL) {
	fprintf(stderr, "Empty input\n");
	return 1;
    }

    suif_proc_iter(argc, argv, do_proc, TRUE);

    if (domains)
	delete [] domains;

    return 0;
}

static void domain_dump(Polyhedron* d)
{
    Polyhedron *p;
    int i;

    for (i = 0, p = d; p; ++i, p = p->next);
    printf("%d\n", i);

    for (p = d; p; p = p->next) {
	int r, c;
        printf("%d %d\n", r=p->NbConstraints, c=p->Dimension+2);
	for (int i=0;i<r;i++) {
	    for (int j=0;j<c;j++)
		value_print(stdout," "P_VALUE_FMT" ", p->Constraint[i][j]);
	    printf("\n");
	}
    }
}

static void program_dump(PDG* pdg)
{
    puts("C");
    puts("0 2");
    puts("0");

    printf("%d\n", pdg->nodes.size());
    for (int i = 0; i < pdg->nodes.size(); ++i) {
	domain_dump(*pdg->nodes[i]->source);
	puts("0 0 0");
    }

    puts("0");

    printf("%d\n", pdg->nodes.size());
    for (int i = 0; i < pdg->nodes.size(); ++i)
	domain_dump(*pdg->nodes[i]->scattering);

    puts("0");
}

static CloogProgram * program_read(char* in, int dump)
{
    FILE *input = fopen(in, "r");
    assert(input);

    PDG * pdg = yaml::Load<PDG>(input);
    fclose(input);
    if (pdg == NULL)
	return NULL;

    if (dump)
	program_dump(pdg);

    cloog_program_builder pb(pdg);

    return pb.p;
}
